from __future__ import annotations

import os
import typing as t
from pathlib import Path

from findpython.providers.base import BaseProvider
from findpython.python import PythonVersion
from findpython.utils import WINDOWS, safe_iter_dir


class RyeProvider(BaseProvider):
    def __init__(self, root: Path) -> None:
        self.root = root

    @classmethod
    def create(cls) -> t.Self | None:
        root = Path(os.getenv("RYE_PY_ROOT", "~/.rye/py")).expanduser()
        return cls(root)

    def find_pythons(self) -> t.Iterable[PythonVersion]:
        if not self.root.exists():
            return
        for child in safe_iter_dir(self.root):
            for intermediate in ("", "install/"):
                if WINDOWS:
                    python_bin = child / (intermediate + "python.exe")
                else:
                    python_bin = child / (intermediate + "bin/python3")
                if python_bin.exists():
                    yield self.version_maker(python_bin, _interpreter=python_bin)
                    break



from pathlib import Path
from unittest.mock import PropertyMock

import pytest
from packaging.version import parse

from findpython.providers import ALL_PROVIDERS, PathProvider
from findpython.python import PythonVersion


class _MockRegistry:
    def __init__(self) -> None:
        self.versions: dict[Path, PythonVersion] = {}

    def add_python(
        self,
        executable,
        version=None,
        architecture="64bit",
        interpreter=None,
        keep_symlink=False,
    ):
        if version is not None:
            version = parse(version)
        executable = Path(executable)
        if interpreter is None:
            interpreter = executable
        executable.parent.mkdir(parents=True, exist_ok=True)
        executable.touch(exist_ok=True)
        executable.chmod(0o744)
        py_ver = PythonVersion(
            executable, version, architecture, interpreter, keep_symlink
        )
        if version is not None:
            py_ver._get_version = lambda: version
        self.versions[executable] = py_ver
        return py_ver

    def version_maker(self, executable, *args, **kwargs):
        return self.versions[executable]


@pytest.fixture()
def mocked_python(tmp_path, monkeypatch):
    mocked = _MockRegistry()
    for python in [
        (tmp_path / "python3.7", "3.7.0"),
        (tmp_path / "python3.8", "3.8.0"),
        (tmp_path / "python3.9", "3.9.0"),
    ]:
        mocked.add_python(*python)
    monkeypatch.setattr(
        "findpython.providers.base.BaseProvider.version_maker", mocked.version_maker
    )
    monkeypatch.setattr(
        "findpython.python.PythonVersion.implementation",
        PropertyMock(return_value="CPython"),
    )
    ALL_PROVIDERS.clear()
    ALL_PROVIDERS["path"] = PathProvider
    monkeypatch.setenv("PATH", str(tmp_path))
    return mocked


@pytest.fixture(params=[False, True])
def switch(request):
    return request.param



# -------------------------------------------------------------------------
# Copyright (c) Steve Dower
# All rights reserved.
#
# Distributed under the terms of the MIT License
# -------------------------------------------------------------------------

__all__ = [
    "open_source",
    "REGISTRY_SOURCE_LM",
    "REGISTRY_SOURCE_LM_WOW6432",
    "REGISTRY_SOURCE_CU",
]

import re
from itertools import count

try:
    import winreg
except ImportError:
    import _winreg as winreg

REGISTRY_SOURCE_LM = 1
REGISTRY_SOURCE_LM_WOW6432 = 2
REGISTRY_SOURCE_CU = 3

_REG_KEY_INFO = {
    REGISTRY_SOURCE_LM: (
        winreg.HKEY_LOCAL_MACHINE,
        r"Software\Python",
        winreg.KEY_WOW64_64KEY,
    ),
    REGISTRY_SOURCE_LM_WOW6432: (
        winreg.HKEY_LOCAL_MACHINE,
        r"Software\Python",
        winreg.KEY_WOW64_32KEY,
    ),
    REGISTRY_SOURCE_CU: (winreg.HKEY_CURRENT_USER, r"Software\Python", 0),
}


def get_value_from_tuple(value, vtype):
    if vtype == winreg.REG_SZ:
        if "\0" in value:
            return value[: value.index("\0")]
        return value
    return None


def join(x, y):
    return x + "\\" + y


_VALID_ATTR = re.compile("^[a-z_]+$")
_VALID_KEY = re.compile("^[A-Za-z]+$")
_KEY_TO_ATTR = re.compile("([A-Z]+[a-z]+)")


class PythonWrappedDict(object):
    @staticmethod
    def _attr_to_key(attr):
        if not attr:
            return ""
        if not _VALID_ATTR.match(attr):
            return attr
        return "".join(c.capitalize() for c in attr.split("_"))

    @staticmethod
    def _key_to_attr(key):
        if not key:
            return ""
        if not _VALID_KEY.match(key):
            return key
        return "_".join(k for k in _KEY_TO_ATTR.split(key) if k).lower()

    def __init__(self, d):
        self._d = d

    def __getattr__(self, attr):
        if attr.startswith("_"):
            return object.__getattribute__(self, attr)

        if attr == "value":
            attr = ""

        key = self._attr_to_key(attr)
        try:
            return self._d[key]
        except Exception:
            pass
        raise AttributeError(attr)

    def __setattr__(self, attr, value):
        if attr.startswith("_"):
            return object.__setattr__(self, attr, value)

        if attr == "value":
            attr = ""
        self._d[self._attr_to_key(attr)] = value

    def __dir__(self):
        k2a = self._key_to_attr
        return list(map(k2a, self._d))

    def _setdefault(self, key, value):
        self._d.setdefault(key, value)

    def _items(self):
        return self._d.items()

    def __repr__(self):
        k2a = self._key_to_attr
        return (
            "info("
            + ", ".join("{}={!r}".format(k2a(k), v) for k, v in self._d.items())
            + ")"
        )


class RegistryAccessor(object):
    def __init__(self, root, subkey, flags):
        self._root = root
        self.subkey = subkey
        _, _, self.name = subkey.rpartition("\\")
        self._flags = flags

    def __iter__(self):
        subkey_names = []
        try:
            with winreg.OpenKeyEx(
                self._root, self.subkey, 0, winreg.KEY_READ | self._flags
            ) as key:
                for i in count():
                    subkey_names.append(winreg.EnumKey(key, i))
        except OSError:
            pass
        return iter(self[k] for k in subkey_names)

    def __getitem__(self, key):
        return RegistryAccessor(self._root, join(self.subkey, key), self._flags)

    def get_value(self, value_name):
        try:
            with winreg.OpenKeyEx(
                self._root, self.subkey, 0, winreg.KEY_READ | self._flags
            ) as key:
                return get_value_from_tuple(*winreg.QueryValueEx(key, value_name))
        except OSError:
            return None

    def get_all_values(self):
        schema = {}
        for subkey in self:
            schema[subkey.name] = subkey.get_all_values()

        key = winreg.OpenKeyEx(self._root, self.subkey, 0, winreg.KEY_READ | self._flags)
        try:
            with key:
                for i in count():
                    vname, value, vtype = winreg.EnumValue(key, i)
                    value = get_value_from_tuple(value, vtype)
                    if value:
                        schema[vname or ""] = value
        except OSError:
            pass

        return PythonWrappedDict(schema)

    def set_value(self, value_name, value):
        with winreg.CreateKeyEx(
            self._root, self.subkey, 0, winreg.KEY_WRITE | self._flags
        ) as key:
            if value is None:
                winreg.DeleteValue(key, value_name)
            elif isinstance(value, str):
                winreg.SetValueEx(key, value_name, 0, winreg.REG_SZ, value)
            else:
                raise TypeError("cannot write {} to registry".format(type(value)))

    def _set_all_values(self, rootkey, name, info, errors):
        with winreg.CreateKeyEx(rootkey, name, 0, winreg.KEY_WRITE | self._flags) as key:
            for k, v in info:
                if isinstance(v, PythonWrappedDict):
                    self._set_all_values(key, k, v._items(), errors)
                elif isinstance(v, dict):
                    self._set_all_values(key, k, v.items(), errors)
                elif v is None:
                    winreg.DeleteValue(key, k)
                elif isinstance(v, str):
                    winreg.SetValueEx(key, k, 0, winreg.REG_SZ, v)
                else:
                    errors.append("cannot write {} to registry".format(type(v)))

    def set_all_values(self, info):
        errors = []
        if isinstance(info, PythonWrappedDict):
            items = info._items()
        elif isinstance(info, dict):
            items = info.items()
        else:
            raise TypeError("info must be a dictionary")

        self._set_all_values(self._root, self.subkey, items, errors)
        if len(errors) == 1:
            raise ValueError(errors[0])
        elif errors:
            raise ValueError(errors)

    def delete(self):
        for k in self:
            k.delete()
        try:
            key = winreg.OpenKeyEx(self._root, None, 0, winreg.KEY_READ | self._flags)
        except OSError:
            return
        with key:
            winreg.DeleteKeyEx(key, self.subkey)


def open_source(registry_source):
    info = _REG_KEY_INFO.get(registry_source)
    if not info:
        raise ValueError("unsupported registry source")
    root, subkey, flags = info
    return RegistryAccessor(root, subkey, flags)



import stat
import sys
from pathlib import Path

import pytest

from findpython import register_provider
from findpython.finder import Finder
from findpython.providers.asdf import AsdfProvider
from findpython.providers.pyenv import PyenvProvider
from findpython.providers.rye import RyeProvider

if sys.platform == "win32":
    pytest.skip("Skip POSIX tests on Windows", allow_module_level=True)


def test_find_python_resolve_symlinks(mocked_python, tmp_path, switch):
    link = Path(tmp_path / "python")
    link.symlink_to(Path(tmp_path / "python3.7"))
    python = mocked_python.add_python(link, "3.7.0")
    finder = Finder(resolve_symlinks=switch)
    all_pythons = finder.find_all()
    assert len(all_pythons) == (3 if switch else 4)
    assert (python in all_pythons) is not switch


def test_find_python_from_asdf(mocked_python, tmp_path, monkeypatch):
    register_provider(AsdfProvider)
    python = mocked_python.add_python(
        tmp_path / ".asdf/installs/python/3.8/bin/python", "3.8.0"
    )
    monkeypatch.setenv("ASDF_DATA_DIR", str(tmp_path / ".asdf"))
    pythons = Finder().find_all(3, 8)
    assert len(pythons) == 2
    assert python in pythons


def test_find_python_exclude_unreadable(mocked_python, tmp_path):
    python = Path(tmp_path / "python3.8")
    python.chmod(python.stat().st_mode & ~stat.S_IRUSR)
    try:
        finder = Finder()
        all_pythons = finder.find_all()
        assert len(all_pythons) == 2, all_pythons
        assert python not in [version.executable for version in all_pythons]
    finally:
        python.chmod(0o744)


def test_find_python_from_provider(mocked_python, tmp_path, monkeypatch):
    register_provider(AsdfProvider)
    register_provider(PyenvProvider)
    python38 = mocked_python.add_python(
        tmp_path / ".asdf/installs/python/3.8/bin/python", "3.8.0"
    )
    python381 = mocked_python.add_python(
        tmp_path / ".pyenv/versions/3.8.1/bin/python", "3.8.1"
    )
    python382 = mocked_python.add_python(
        tmp_path / ".asdf/installs/python/3.8.2/bin/python", "3.8.2"
    )
    monkeypatch.setenv("ASDF_DATA_DIR", str(tmp_path / ".asdf"))
    monkeypatch.setenv("PYENV_ROOT", str(tmp_path / ".pyenv"))

    pythons = Finder(selected_providers=["pyenv", "asdf"]).find_all(3, 8)
    assert len(pythons) == 3
    assert python38 in pythons
    assert python381 in pythons
    assert python382 in pythons

    asdf_pythons = Finder(selected_providers=["asdf"]).find_all(3, 8)
    assert len(asdf_pythons) == 2
    assert python38 in asdf_pythons
    assert python382 in asdf_pythons

    pyenv_pythons = Finder(selected_providers=["pyenv"]).find_all(3, 8)
    assert len(pyenv_pythons) == 1
    assert python381 in pyenv_pythons


def test_find_python_from_rye_provider(mocked_python, tmp_path, monkeypatch):
    python310 = mocked_python.add_python(
        tmp_path / ".rye/py/cpython@3.10.9/install/bin/python3", "3.10.9"
    )
    python311 = mocked_python.add_python(
        tmp_path / ".rye/py/cpython@3.11.8/bin/python3", "3.11.8"
    )
    monkeypatch.setenv("HOME", str(tmp_path))

    register_provider(RyeProvider)
    find_310 = Finder(selected_providers=["rye"]).find_all(3, 10)
    assert python310 in find_310

    find_311 = Finder(selected_providers=["rye"]).find_all(3, 11)
    assert python311 in find_311



import pytest

from findpython.utils import WINDOWS, looks_like_python

matrix = [
    ("python", True),
    ("python3", True),
    ("python38", True),
    ("python3.8", True),
    ("python3.10", True),
    ("python310", True),
    ("python3.6m", True),
    ("python3.6.8m", False),
    ("anaconda3.3", True),
    ("python-3.8.10", False),
    ("unknown-2.0.0", False),
    ("python3.8.unknown", False),
    ("python38.bat", WINDOWS),
    ("python38.exe", WINDOWS),
    ("python38.sh", not WINDOWS),
    ("python38.csh", not WINDOWS),
]


@pytest.mark.parametrize("name, expected", matrix)
def test_looks_like_python(name, expected):
    assert looks_like_python(name) == expected



"""
This package contains all the providers for the pythonfinder module.
"""
from __future__ import annotations

from findpython.providers.asdf import AsdfProvider
from findpython.providers.base import BaseProvider
from findpython.providers.macos import MacOSProvider
from findpython.providers.path import PathProvider
from findpython.providers.pyenv import PyenvProvider
from findpython.providers.rye import RyeProvider
from findpython.providers.winreg import WinregProvider

_providers: list[type[BaseProvider]] = [
    # General:
    PathProvider,
    # Tool Specific:
    AsdfProvider,
    PyenvProvider,
    RyeProvider,
    # Windows only:
    WinregProvider,
    # MacOS only:
    MacOSProvider,
]

ALL_PROVIDERS = {cls.name(): cls for cls in _providers}

__all__ = [cls.__name__ for cls in _providers] + ["ALL_PROVIDERS", "BaseProvider"]



import os
from pathlib import Path

import pytest
from packaging.version import Version

from findpython import Finder, register_provider
from findpython.providers.pyenv import PyenvProvider


def test_find_pythons(mocked_python, tmp_path):
    finder = Finder()
    all_pythons = finder.find_all()
    assert len(all_pythons) == 3
    assert all_pythons[0].executable == Path(tmp_path / "python3.9")
    assert all_pythons[0].version == Version("3.9.0")
    assert all_pythons[1].executable == Path(tmp_path / "python3.8")
    assert all_pythons[1].version == Version("3.8.0")
    assert all_pythons[2].executable == Path(tmp_path / "python3.7")
    assert all_pythons[2].version == Version("3.7.0")


def test_find_python_by_version(mocked_python, tmp_path):
    finder = Finder()
    python = finder.find(3, 8)
    assert python.executable == Path(tmp_path / "python3.8")
    assert python.version == Version("3.8.0")

    assert finder.find("3.8") == python
    assert finder.find("3.8.0") == python
    assert finder.find("python3.8") == python


def test_find_python_by_version_not_found(mocked_python, tmp_path):
    finder = Finder()
    python = finder.find(3, 10)
    assert python is None


def test_find_python_by_architecture(mocked_python, tmp_path):
    python = mocked_python.add_python(
        tmp_path / "python38", "3.8.0", architecture="32bit"
    )
    finder = Finder()
    assert finder.find(3, 8, architecture="32bit") == python
    assert finder.find(3, 8, architecture="64bit").executable == tmp_path / "python3.8"


def test_find_python_with_prerelease(mocked_python, tmp_path):
    python = mocked_python.add_python(tmp_path / "python3.10", "3.10.0.a1")
    finder = Finder()
    assert python == finder.find(pre=True)


def test_find_python_with_devrelease(mocked_python, tmp_path):
    python = mocked_python.add_python(tmp_path / "python3.10", "3.10.0.dev1")
    finder = Finder()
    assert python == finder.find(dev=True)


def test_find_python_with_non_existing_path(mocked_python, monkeypatch):
    monkeypatch.setenv("PATH", "/non/existing/path" + os.pathsep + os.environ["PATH"])
    finder = Finder()
    all_pythons = finder.find_all()
    assert len(all_pythons) == 3


def test_find_python_exclude_invalid(mocked_python, tmp_path):
    python = mocked_python.add_python(tmp_path / "python3.10")
    finder = Finder()
    all_pythons = finder.find_all()
    assert len(all_pythons) == 3
    assert python not in all_pythons


def test_find_python_deduplicate_same_file(mocked_python, tmp_path, switch):
    for i, python in enumerate(mocked_python.versions):
        python.write_bytes(str(i).encode())

    new_python = mocked_python.add_python(tmp_path / "python3", "3.9.0")
    new_python.executable.write_bytes(b"0")

    finder = Finder(no_same_file=switch)
    all_pythons = finder.find_all()
    assert len(all_pythons) == (3 if switch else 4)
    assert (new_python in all_pythons) is not switch


@pytest.mark.skipif(os.name == "nt", reason="Not supported on Windows")
def test_find_python_deduplicate_symlinks(mocked_python, tmp_path):
    python = mocked_python.add_python(tmp_path / "python3.9", "3.9.0")
    (tmp_path / "python3").symlink_to(python.executable)
    symlink1 = mocked_python.add_python(tmp_path / "python3", "3.9.0")
    (tmp_path / "python").symlink_to(python.executable)
    symlink2 = mocked_python.add_python(tmp_path / "python", "3.9.0", keep_symlink=True)
    finder = Finder(resolve_symlinks=True)
    all_pythons = finder.find_all()
    assert python in all_pythons
    assert symlink1 not in all_pythons
    assert symlink2 in all_pythons


def test_find_python_deduplicate_same_interpreter(mocked_python, tmp_path, switch):
    if os.name == "nt":
        suffix = ".bat"
    else:
        suffix = ".sh"
    python = mocked_python.add_python(
        tmp_path / f"python{suffix}", "3.9.0", interpreter=tmp_path / "python3.9"
    )

    finder = Finder(no_same_interpreter=switch)
    all_pythons = finder.find_all()
    assert len(all_pythons) == (3 if switch else 4)
    assert (python in all_pythons) is not switch


def test_find_python_from_pyenv(mocked_python, tmp_path, monkeypatch):
    register_provider(PyenvProvider)
    python = mocked_python.add_python(
        tmp_path / ".pyenv/versions/3.8/bin/python", "3.8.0"
    )
    monkeypatch.setenv("PYENV_ROOT", str(tmp_path / ".pyenv"))
    pythons = Finder().find_all(3, 8)
    assert len(pythons) == 2
    assert python in pythons


def test_find_python_skips_empty_pyenv(mocked_python, tmp_path, monkeypatch):
    register_provider(PyenvProvider)
    pyenv_path = Path(tmp_path / ".pyenv")
    pyenv_path.mkdir()
    monkeypatch.setenv("PYENV_ROOT", str(pyenv_path))
    all_pythons = Finder().find_all()
    assert len(all_pythons) == 3



from __future__ import annotations

import abc
import logging
import typing as t
from pathlib import Path

from findpython.python import PythonVersion
from findpython.utils import path_is_python, safe_iter_dir

logger = logging.getLogger("findpython")


class BaseProvider(metaclass=abc.ABCMeta):
    """The base class for python providers"""

    version_maker: t.Callable[..., PythonVersion] = PythonVersion

    @classmethod
    def name(cls) -> str:
        """Configuration name for this provider.

        By default, the lowercase class name with 'provider' removed.
        """
        self_name = cls.__name__.lower()
        if self_name.endswith("provider"):
            self_name = self_name[: -len("provider")]
        return self_name

    @classmethod
    @abc.abstractmethod
    def create(cls) -> t.Self | None:
        """Return an instance of the provider or None if it is not available"""
        pass

    @abc.abstractmethod
    def find_pythons(self) -> t.Iterable[PythonVersion]:
        """Return the python versions found by the provider"""
        pass

    @classmethod
    def find_pythons_from_path(
        cls, path: Path, as_interpreter: bool = False
    ) -> t.Iterable[PythonVersion]:
        """A general helper method to return pythons under a given path.

        :param path: The path to search for pythons
        :param as_interpreter: Use the path as the interpreter path.
            If the pythons might be a wrapper script, don't set this to True.
        :returns: An iterable of PythonVersion objects
        """
        return (
            cls.version_maker(
                child.absolute(),
                _interpreter=child.absolute() if as_interpreter else None,
            )
            for child in safe_iter_dir(path)
            if path_is_python(child)
        )



from __future__ import annotations

import typing as t
from pathlib import Path

from findpython.providers.base import BaseProvider
from findpython.python import PythonVersion


class MacOSProvider(BaseProvider):
    """A provider that finds python from macos typical install base
    with python.org installer.
    """

    INSTALL_BASE = Path("/Library/Frameworks/Python.framework/Versions/")

    @classmethod
    def create(cls) -> t.Self | None:
        if not cls.INSTALL_BASE.exists():
            return None
        return cls()

    def find_pythons(self) -> t.Iterable[PythonVersion]:
        for version in self.INSTALL_BASE.iterdir():
            if version.is_dir():
                yield from self.find_pythons_from_path(version / "bin", True)



from __future__ import annotations

import logging
import sys
from argparse import ArgumentParser

from findpython import Finder
from findpython.__version__ import __version__

logger = logging.getLogger("findpython")


def setup_logger(level: int = logging.DEBUG) -> None:
    """
    Setup the logger.
    """
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("%(name)s-%(levelname)s: %(message)s"))
    logger.addHandler(handler)
    logger.setLevel(level)


def split_str(value: str) -> list[str]:
    return value.split(",")


def cli(argv: list[str] | None = None) -> int:
    """
    Command line interface for findpython.
    """
    parser = ArgumentParser(
        "findpython", description="A utility to find python versions on your system"
    )
    parser.add_argument(
        "-V", "--version", action="version", version=f"%(prog)s {__version__}"
    )
    parser.add_argument(
        "-a", "--all", action="store_true", help="Show all matching python versions"
    )
    parser.add_argument(
        "--resolve-symlink", action="store_true", help="Resolve all symlinks"
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Verbose output")
    parser.add_argument(
        "--no-same-file",
        action="store_true",
        help="Eliminate the duplicated results with the same file contents",
    )
    parser.add_argument(
        "--no-same-python",
        action="store_true",
        help="Eliminate the duplicated results with the same sys.executable",
    )
    parser.add_argument(
        "--pre", "--prereleases", action="store_true", help="Allow prereleases"
    )
    parser.add_argument("--providers", type=split_str, help="Select provider(s) to use")
    parser.add_argument("version_spec", nargs="?", help="Python version spec or name")

    args = parser.parse_args(argv)
    if args.verbose:
        setup_logger()

    finder = Finder(
        resolve_symlinks=args.resolve_symlink,
        no_same_file=args.no_same_file,
        selected_providers=args.providers,
    )
    if args.all:
        find_func = finder.find_all
    else:
        find_func = finder.find  # type: ignore[assignment]

    python_versions = find_func(args.version_spec, allow_prereleases=args.pre)
    if not python_versions:
        print("No matching python version found", file=sys.stderr)
        return 1
    if not isinstance(python_versions, list):
        python_versions = [python_versions]
    print("Found matching python versions:", file=sys.stderr)
    for python_version in python_versions:
        print(python_version)
    return 0


def main() -> None:
    """
    Main function.
    """
    sys.exit(cli())


if __name__ == "__main__":
    main()



from findpython.__main__ import cli


def test_cli_find_pythons(mocked_python, capsys):
    retcode = cli(["--all"])
    assert retcode == 0
    out, _ = capsys.readouterr()
    lines = out.strip().splitlines()
    for version, line in zip(("3.9", "3.8", "3.7"), lines):
        assert line.lstrip().startswith(f"CPython@{version}.0")


def test_cli_find_python_by_version(mocked_python, capsys, tmp_path):
    retcode = cli(["3.8"])
    assert retcode == 0
    out, _ = capsys.readouterr()
    line = out.strip()
    assert line.startswith("CPython@3.8.0")
    assert line.endswith(str(tmp_path / "python3.8"))



from __future__ import annotations

import os
import typing as t
from pathlib import Path

from findpython.providers.base import BaseProvider
from findpython.python import PythonVersion


class AsdfProvider(BaseProvider):
    """A provider that finds python installed with asdf"""

    def __init__(self, root: Path) -> None:
        self.root = root

    @classmethod
    def create(cls) -> t.Self | None:
        asdf_root = os.path.expanduser(
            os.path.expandvars(os.getenv("ASDF_DATA_DIR", "~/.asdf"))
        )
        if not os.path.exists(asdf_root):
            return None
        return cls(Path(asdf_root))

    def find_pythons(self) -> t.Iterable[PythonVersion]:
        python_dir = self.root / "installs/python"
        if not python_dir.exists():
            return
        for version in python_dir.iterdir():
            if version.is_dir():
                bindir = version / "bin"
                if not bindir.exists():
                    bindir = version
                yield from self.find_pythons_from_path(bindir, True)



from __future__ import annotations

import logging
import operator
from typing import Callable, Iterable

from findpython.providers import ALL_PROVIDERS, BaseProvider
from findpython.python import PythonVersion
from findpython.utils import get_suffix_preference, parse_major

logger = logging.getLogger("findpython")


class Finder:
    """Find python versions on the system.

    :param resolve_symlinks: Whether to resolve symlinks.
    :param no_same_file: Whether to deduplicate with the python executable content.
    :param no_same_interpreter: Whether to deduplicate with the python executable path.
    """

    def __init__(
        self,
        resolve_symlinks: bool = False,
        no_same_file: bool = False,
        no_same_interpreter: bool = False,
        selected_providers: list[str] | None = None,
    ) -> None:
        self.resolve_symlinks = resolve_symlinks
        self.no_same_file = no_same_file
        self.no_same_interpreter = no_same_interpreter
        self._providers = self.setup_providers(selected_providers)

    def setup_providers(
        self,
        selected_providers: list[str] | None = None,
    ) -> list[BaseProvider]:
        providers: list[BaseProvider] = []
        allowed_providers = ALL_PROVIDERS
        if selected_providers is not None:
            allowed_providers = {name: ALL_PROVIDERS[name] for name in selected_providers}
        for provider_class in allowed_providers.values():
            provider = provider_class.create()
            if provider is None:
                logger.debug("Provider %s is not available", provider_class.__name__)
            else:
                providers.append(provider)
        return providers

    def add_provider(self, provider: BaseProvider, pos: int | None = None) -> None:
        """Add provider to the provider list.
        If pos is given, it will be inserted at the given position.
        """
        if pos is not None:
            self._providers.insert(pos, provider)
        else:
            self._providers.append(provider)

    def find_all(
        self,
        major: int | str | None = None,
        minor: int | None = None,
        patch: int | None = None,
        pre: bool | None = None,
        dev: bool | None = None,
        name: str | None = None,
        architecture: str | None = None,
        allow_prereleases: bool = False,
        implementation: str | None = None,
    ) -> list[PythonVersion]:
        """
        Return all Python versions matching the given version criteria.

        :param major: The major version or the version string or the name to match.
        :param minor: The minor version to match.
        :param patch: The micro version to match.
        :param pre: Whether the python is a prerelease.
        :param dev: Whether the python is a devrelease.
        :param name: The name of the python.
        :param architecture: The architecture of the python.
        :param allow_prereleases: Whether to allow prereleases.
        :param implementation: The implementation of the python. E.g. "cpython", "pypy".
        :return: a list of PythonVersion objects
        """
        if allow_prereleases and (pre is False or dev is False):
            raise ValueError(
                "If allow_prereleases is True, pre and dev must not be False."
            )
        if isinstance(major, str):
            if any(v is not None for v in (minor, patch, pre, dev, name)):
                raise ValueError(
                    "If major is a string, minor, patch, pre, dev and name "
                    "must not be specified."
                )
            version_dict = parse_major(major)
            if version_dict is not None:
                major = version_dict["major"]
                minor = version_dict["minor"]
                patch = version_dict["patch"]
                pre = version_dict["pre"]
                dev = version_dict["dev"]
                if allow_prereleases:
                    pre = pre or None
                    dev = dev or None
                architecture = version_dict["architecture"]
                implementation = version_dict["implementation"]
            else:
                name, major = major, None

        version_matcher = operator.methodcaller(
            "matches",
            major,
            minor,
            patch,
            pre,
            dev,
            name,
            architecture,
            implementation,
        )
        # Deduplicate with the python executable path
        matched_python = set(self._find_all_python_versions())
        return self._dedup(matched_python, version_matcher)

    def find(
        self,
        major: int | str | None = None,
        minor: int | None = None,
        patch: int | None = None,
        pre: bool | None = None,
        dev: bool | None = None,
        name: str | None = None,
        architecture: str | None = None,
        allow_prereleases: bool = False,
        implementation: str | None = None,
    ) -> PythonVersion | None:
        """
        Return the Python version that is closest to the given version criteria.

        :param major: The major version or the version string or the name to match.
        :param minor: The minor version to match.
        :param patch: The micro version to match.
        :param pre: Whether the python is a prerelease.
        :param dev: Whether the python is a devrelease.
        :param name: The name of the python.
        :param architecture: The architecture of the python.
        :param allow_prereleases: Whether to allow prereleases.
        :param implementation: The implementation of the python. E.g. "cpython", "pypy".
        :return: a Python object or None
        """
        return next(
            iter(
                self.find_all(
                    major,
                    minor,
                    patch,
                    pre,
                    dev,
                    name,
                    architecture,
                    allow_prereleases,
                    implementation,
                )
            ),
            None,
        )

    def _find_all_python_versions(self) -> Iterable[PythonVersion]:
        """Find all python versions on the system."""
        for provider in self._providers:
            yield from provider.find_pythons()

    def _dedup(
        self,
        python_versions: Iterable[PythonVersion],
        version_matcher: Callable[[PythonVersion], bool],
    ) -> list[PythonVersion]:
        def dedup_key(python_version: PythonVersion) -> str:
            if self.no_same_interpreter:
                return python_version.interpreter.as_posix()
            if self.no_same_file:
                return python_version.binary_hash()
            if self.resolve_symlinks and not python_version.keep_symlink:
                return python_version.real_path.as_posix()
            return python_version.executable.as_posix()

        def sort_key(python_version: PythonVersion) -> tuple[int, int, int]:
            return (
                python_version.executable.is_symlink(),
                get_suffix_preference(python_version.name),
                -len(python_version.executable.as_posix()),
            )

        result: dict[str, PythonVersion] = {}

        for python_version in sorted(python_versions, key=sort_key):
            key = dedup_key(python_version)
            if (
                key not in result
                and python_version.is_valid()
                and version_matcher(python_version)
            ):
                result[key] = python_version
        return sorted(result.values(), reverse=True)



from __future__ import annotations

import os
import typing as t
from dataclasses import dataclass
from pathlib import Path

from findpython.providers.base import BaseProvider
from findpython.python import PythonVersion


@dataclass
class PathProvider(BaseProvider):
    """A provider that finds Python from PATH env."""

    paths: list[Path]

    @classmethod
    def create(cls) -> t.Self | None:
        paths = [Path(path) for path in os.getenv("PATH", "").split(os.pathsep) if path]
        return cls(paths)

    def find_pythons(self) -> t.Iterable[PythonVersion]:
        for path in self.paths:
            yield from self.find_pythons_from_path(path)
